const bsv = require('babbage-bsv')
const { getPaymentPrivateKey, getPaymentAddress } = require('sendover')
const makeCreateAuthriteRequest = require('./utils/makeCreateAuthriteRequest')

/**
 * @typedef {Object} GetTxWithOutputsOutput
 * @property {String} script The hex representing the locking script of the output
 * @property {Number} satoshis The number of satoshis to put in the output
 */

/**
 * @typedef {Object} GetTxWithOutputsResult
 * @property {String} rawTx The serialized, signed transaction that is ready for broadcast
 * @property {String} txid The txid for the new transaction
 * @property {String} referenceNumber The reference number that should now be provided back to `processTransaction (or `updateTransactionStatus`)
 * @property {Object} inputs This is the fully-formed `inputs` field of this transaction, as per the SPV Envelope specification.
 * @property {Number} amount The amount of the transaction
 */

module.exports = async ({
  authriteClient,
  config,
  processTransaction,
  outputs = [],
  feePerKb = 110,
  labels,
  inputs = {},
  autoProcess = true,
  recipient,
  note
} = {}) => {
  const createAuthriteRequest = makeCreateAuthriteRequest({
    authriteClient,
    config
  })

  const createResult = await createAuthriteRequest('createTransaction', {
    method: 'POST',
    body: {
      inputs: Object.fromEntries(Object.entries(inputs).map(([k, v]) => ([k, {
        ...v, // Calculate unlockingScriptLength from unlockingScript
        outputsToRedeem: v.outputsToRedeem.map(x => ({
          unlockingScriptLength: x.unlockingScript.length / 2,
          index: x.index
        }))
      }]))),
      outputs,
      fee: {
        model: 'sat/kb',
        value: feePerKb
      },
      labels,
      note,
      recipient
    }
  })
  if (createResult.status === 'error') {
    return createResult
  }
  const {
    inputs: dojoInputs,
    outputs: txOutputs,
    referenceNumber,
    derivationPrefix,
    paymailHandle
  } = createResult
  const tx = new bsv.Transaction()
  const outputMap = {}

  txOutputs.forEach((out, i) => {
    if (out.providedBy === 'dojo' && out.purpose === 'change') {
      // Derive a change output script
      // Get derivation invoice data
      const derivationSuffix = out.derivationSuffix
      outputMap[derivationSuffix] = i
      const invoiceNumber = paymailHandle
        ? `2-3241645161d8-${paymailHandle} ${derivationPrefix} ${derivationSuffix}`
        : `2-3241645161d8-${derivationPrefix} ${derivationSuffix}`
      // Derive the public key used for creating the output script
      const derivedAddress = getPaymentAddress({
        senderPrivateKey: authriteClient.clientPrivateKey,
        recipientPublicKey: authriteClient.clientPublicKey,
        invoiceNumber,
        returnType: 'address'
      })
      // Create an output script that can only be unlocked with the corresponding derived private key
      tx.addOutput(new bsv.Transaction.Output({
        script: new bsv.Script(
          bsv.Script.fromAddress(derivedAddress)
        ),
        satoshis: out.satoshis
      }))
    } else {
      tx.addOutput(new bsv.Transaction.Output({
        script: new bsv.Script(out.script),
        satoshis: out.satoshis
      }))
    }
  })

  const getIndex = o => {
    if (typeof o === 'object') {
      return o.index
    } else {
      return o
    }
  }

  // Add inputs, and sum input amounts
  let totalInputs = 0
  for (const [inputTXID, input] of Object.entries(dojoInputs)) {
    const t = new bsv.Transaction(input.rawTx)
    for (const o of input.outputsToRedeem) {
      tx.from(bsv.Transaction.UnspentOutput({
        txid: inputTXID,
        outputIndex: getIndex(o),
        // scruptPubKey a.k.a. lockingScript or outputScript
        // (whereas scriptSig a.k.a. unlockingScript or inputScript)
        scriptPubKey: t.outputs[getIndex(o)].script,
        satoshis: t.outputs[getIndex(o)].satoshis
      }))
      if ( // All foreign input scripts are added
        inputs[t.id] &&
        inputs[t.id].outputsToRedeem.some(x => x.index === getIndex(o))
      ) {
        tx.inputs[tx.inputs.length - 1].setScript(bsv.Script.fromHex(
          inputs[t.id].outputsToRedeem.find(x => x.index === getIndex(o)).unlockingScript
        ))
        // This overrides an abstract method on custom input types, indicating that the entire unlocking script is already present for this foreign input, and no new signatures are ever needed.
        tx.inputs[tx.inputs.length - 1].getSignatures = () => ([])
      } else { // All non-foreign inputs are summed
        totalInputs += t.outputs[getIndex(o)].satoshis
      }
    }
  }

  //  Sign inputs using type42 derived key
  for (const input of Object.values(dojoInputs)) {
    for (const o of input.outputsToRedeem) {
      if (input.instructions && input.instructions[getIndex(o)]) {
        const instructions = input.instructions[getIndex(o)]

        // Make sure the transaction type is supported
        if (instructions.type !== 'P2PKH') {
          throw new Error(
            `The server is specifying ${instructions.type}, which is not a supported unlocking script type. Supported type: P2PKH`
          )
        }

        // Get derivation invoice data
        const paymailHandle = instructions.paymailHandle
        const derivationPrefix = instructions.derivationPrefix
        const derivationSuffix = instructions.derivationSuffix

        const invoiceNumber = paymailHandle
          ? `2-3241645161d8-${paymailHandle} ${derivationPrefix} ${derivationSuffix}`
          : `2-3241645161d8-${derivationPrefix} ${derivationSuffix}`
        // Derive the key used to unlock funds
        const derivedPrivateKey = getPaymentPrivateKey({
          recipientPrivateKey: authriteClient.clientPrivateKey,
          senderPublicKey: instructions.senderIdentityKey,
          invoiceNumber
        })
        tx.sign(bsv.PrivateKey.fromWIF(derivedPrivateKey))
      }
    }
  }

  // The amount is the total of inputs minus change outputs
  // Note that the amount can be negative when we are redeeming more inputs than we are spending
  const amount = totalInputs - txOutputs.filter(x => x.purpose === 'change')
    .reduce((acc, el) => acc + el.satoshis, 0)

  // The inputs are sanitized to remove instructions
  const sanitizedInputs = Object.fromEntries(
    Object.entries(dojoInputs).map(([k, v]) => ([k, {
      inputs: v.inputs,
      mapiResponses: v.mapiResponses,
      proof: v.proof,
      rawTx: v.rawTx
    }]))
  )

  // Return an SPV Envelope
  if (!autoProcess) {
    return {
      rawTx: tx.uncheckedSerialize(),
      txid: tx.id,
      referenceNumber,
      amount,
      inputs: sanitizedInputs,
      outputMap
    }
  } else {
    const hex = tx.uncheckedSerialize()
    const { mapiResponses, note } = await processTransaction({
      submittedTransaction: hex,
      reference: referenceNumber,
      outputMap,
      inputs: sanitizedInputs
    })
    return {
      rawTx: hex,
      txid: tx.id,
      mapiResponses,
      note,
      amount,
      inputs: sanitizedInputs
    }
  }
}
