const { CONFIG } = require('./defaults')
const { Authrite } = require('authrite-js')
const boomerang = require('boomerang-http')
const makeCreateAuthriteRequest = require('./utils/makeCreateAuthriteRequest')

const _getTransactionWithOutputs = require('./getTransactionWithOutputs')
const _processPendingTransactions = require('./processPendingTransactions')
const _processIncomingTransaction = require('./utils/processIncomingTransaction')

const bsv = require('babbage-bsv')
const axios = require('axios')
const WhatsOnChain = require('whatsonchain')
const hashwrap = require('hash-wrap')

/**
 * Get Object Type Definitions
 * @typedef {import('./utils/objects').Avatar} Avatar
 * @typedef {import('./utils/objects').GetTransactionsTx} GetTransactionsTx
 * @typedef {import('./utils/objects').GetTransactionsResult} GetTransactionsResult
 * @typedef {import('./utils/objects').GetPendingTransactionsTx} GetPendingTransactionsTx
 * @typedef {import('./utils/objects').GetTxWithOutputsOutput} GetTxWithOutputsOutput
 * @typedef {import('./utils/objects').GetTxWithOutputsResult} GetTxWithOutputsResult
 * @typedef {import('./utils/objects').TxRedeemableOutput} TxRedeemableOutput
 * @typedef {import('./utils/objects').TxOutput} TxOutput
 * @typedef {import('./utils/objects').InputTemplate} InputTemplate
 * @typedef {import('./utils/objects').OutputTemplate} OutputTemplate
 * @typedef {import('./utils/objects').TransactionTemplate} TransactionTemplate
 * @typedef {import('./utils/objects').TxInputEnvelope} TxInputEnvelope
 * @typedef {import('./utils/objects').TransactionOutputDescriptor} TransactionOutputDescriptor
 */

class Ninja {
  /**
 * A client for creating, signing, and delivering Bitcoin transactions
 * @param {Object} obj All parameters are given in an object
 * @param {Object} [obj.privateKey] (Optional) The client private key to use for the authrite client. If none is provided, the BabbageSDK will be used as the signing strategy.
 * @param {Object} [obj.config] (Optional) Config object, see config section
 * @param {Object} [obj.taalApiKeys] (Optional) Taal API Keys to be used for both testnet and mainnet.
 * @constructor
 */
  constructor ({ privateKey, config = CONFIG, taalApiKeys = { test: 'testnet_ba132cc4d5b2ebde7ed0ee0f6ee3f678', main: 'mainnet_6c8f8c37afd5c45e09f62d083288a181' } } = {}) {
    this.config = config
    this.authriteClient = new Authrite({
      clientPrivateKey: privateKey
    })
    this.taalApiKeys = taalApiKeys
  }

  /**
   * Wraps the Authrite.request method to provide additional functionality
   * @param {String} feature The request destination route
   * @param {Object} fetchConfig Any configuration including the request body
   * @returns {Object} JSON object containing the result of the request
   */
  async createAuthriteRequest (feature, fetchConfig) {
    return makeCreateAuthriteRequest({
      authriteClient: this.authriteClient,
      config: this.config
    })(feature, fetchConfig)
  }

  /**
   * Returns the current Paymail handle
   * @returns {Promise<String>} The Paymail handle
   */
  async getPaymail () {
    const result = await this.createAuthriteRequest('getCurrentPaymails', {
      method: 'POST',
      body: {}
    })
    return result.aliases[0]
  }

  /**
   * Changes the Paymail handle of the user. NOTE that the old handle will be available for others to use. NOTE that to prevent span, you may only do this if there is at least one unspent output under Dojo management.
   * @param {Object} obj All parameters are given in an object
   * @param {String} obj.paymail A new Paymail handle
   * @returns {Promise<Object>} A success object with `status: "success"`
   */
  async setPaymail ({ paymail }) {
    return await this.createAuthriteRequest('changePaymail', {
      method: 'POST',
      body: { paymail }
    })
  }

  /**
   * Returns the name and photo URL of the user
   * @returns {Promise<Avatar>} The avatar of the user
   */
  async getAvatar () {
    const paymail = await this.getPaymail()
    const [alias, domain] = paymail.split('@')
    const result = await boomerang(
      'GET',
      `${this.config.dojoURL}/getAvatarForPaymail/${alias}@${domain}`
    )
    if (typeof result === 'object' && result.status === 'error') {
      throw new Error(
        `${result.code}: ${result.description}`
      )
    }
    return result
  }

  /**
   * Sets a new name and photo URL
   * @param {Object} obj All parameters are given in an object
   * @param {String} obj.name A new name
   * @param {String} obj.photoURL A new UHRP or HTTPS URL to a photo of the user
   * @returns {Promise<Object>} A success object with `status: "success"`
   */
  async setAvatar ({ name, photoURL }) {
    const paymail = await this.getPaymail()
    return this.createAuthriteRequest('setNameAndPhotoURL', {
      method: 'POST',
      body: { name, photoURL, paymail }
    })
  }

  /**
   * Returns a set of transactions that match the criteria
   * @param {Object} obj All parameters are given in an object
   * @param {String} [obj.involving] Only include transactions sent to or received from this Paymail handle
   * @param {String} [obj.status] If provided, only transactions with the given status will be returned
   * @param {String} [obj.label] Only include transactions affixed with this label
   * @param {Number} [obj.limit=25] Include only this maximum number of transactions in the set
   * @param {Number} [obj.offset=0] Include transactions offset by this number in the full set
   * @param {String} [obj.orderBy=newest-first] Sort order, can be `newest-first` or `oldest-first`
   * @returns {Promise<GetTransactionsResult>} The transactions and the total size of the set
   */
  async getTransactions ({
    limit = 25,
    offset = 0,
    involving,
    orderBy = 'newest-first',
    label,
    status,
    referenceNumber
  } = {}) {
    return await this.createAuthriteRequest('getTransactions', {
      method: 'POST',
      body: { limit, offset, involving, orderBy, label, status, referenceNumber }
    })
  }

  /**
   * Returns a set of all transactions that need to be signed and submitted, or canceled
   * @returns {Promise<Array<GetPendingTransactionsTx>>} The array of pending transactions
   */
  async getPendingTransactions ({ referenceNumber } = {}) {
    return await this.createAuthriteRequest('getPendingTransactions', {
      method: 'POST',
      body: { referenceNumber }
    })
  }

  /**
   * Returns the total of the amounts of transactions that fall under certain criteria
   * @param {Object} obj All parameters are given in an object
   * @param {String} [obj.involving] The Paymail handle of a user who must be involved with the transactions
   * @param {String} [obj.label] The label that must be affixed to the transactions
   * @param {String} obj.direction Must be either "incoming" or "outgoing", transactions in only one direction can be totaled
   * @param {Number} [obj.startTime] The earlies time of the transactions to include, an epoch timestamp in seconds
   * @param {Number} [obj.endTime] The latest time of the transactions to include, an epoch timestamp in seconds
   * @returns {Promise<Object>} An object containing `total`, a number of satoshis
   */
  async getTotalOfAmounts ({
    involving,
    label,
    direction,
    startTime,
    endTime
  } = {}) {
    return await this.createAuthriteRequest('getTotalOfAmounts', {
      method: 'POST',
      body: { involving, label, direction, startTime, endTime }
    })
  }

  /**
   * Returns the total of unspent outputs
   * @returns {Promise<Object>} An object containing `total`, a non-negative number of satoshis
   */
  async getTotalValue () {
    return await this.createAuthriteRequest('getTotalOfUnspentOutputs', {
      method: 'POST',
      body: {}
    })
  }

  /**
 * Creates and signs a transaction with specified outputs, so that it can be processed with `processTransaction`. This is a higher-level wrapper around `createTransaction` so that you do not need to manually handle signing, when you are not providing any non-Dojo inputs.
 *
 * Use this by default, and fall back to `createTransaction` if you need more customization.
 *
 * @param {Object} obj All parameters are given in an object
 * @param {Authrite} obj.authriteClient An API for invoking mutually authenticated requests
 * @param {Array<GetTxWithOutputsOutput>} [obj.outputs=[]] A set of outputs to include, each with `script` and `satoshis`.
 * @param {Number} [obj.feePerKb=110] The number of satoshis to pay per KB of block space used by this transaction.
 * @param {Array<String>} [obj.labels=[]] A set of label strings to affix to the transaction
 * @param {Object} [obj.inputs={}] Input scripts to spend as part of this transaction. This is an object whose keys are TXIDs and whose values are Everett-style transaction envelopes that contain an additional field called `outputsToRedeem`. This is an array of objects, each containing `index` and `unlockingScript` properties. The `index` property is the output number in the transaction you are spending, and `unlockingScript` is the hex scriptcode that unlocks the satoshis. Note that you should create any signatures with `SIGHASH_NONE | ANYONECANPAY` or similar so that the additional Dojo outputs can be added afterward without invalidating your signature.
 * @param {Boolean} [obj.autoProcess=true] Whether the transaction should be processed automatically with processTransaction. Note that this will return `mapiResponses` and `note` instead of referenceNumber
 * @param {String} [obj.recipient] Paymail recipient for transaction
 * @param {String} [obj.note] A note about the transaction
 * @returns {Promise<GetTxWithOutputsResult>} The serialized transaction, inputs, reference number and amount
*/
  getTransactionWithOutputs ({
    outputs = [],
    feePerKb = 110,
    labels,
    inputs = {},
    autoProcess = true,
    note,
    recipient
  } = {}) {
    return _getTransactionWithOutputs({
      authriteClient: this.authriteClient,
      config: this.config,
      processTransaction: this.processTransaction.bind(this),
      outputs,
      note,
      recipient,
      feePerKb,
      labels,
      inputs,
      autoProcess
    })
  }

  /**
   * Creates a new transaction that must be processed with `processTransaction`
   * after you sign it
   *
   * @param {Object} obj All parameters are given in an object
   * @param {Object<TxInputEnvelope>} [obj.inputs] Specify any additional inputs
   *                 to the transaction (if any) that are not to be provided by
   *                 the
   *                 Dojo. If you do not provide inputs here, or if they are
   *                 insufficient, Dojo will select additional inputs for you to
   *                 sign. To control this input selection behavior, see the
   *                 `inputSelection` parameter. This `inputs` parameter is an
   *                 object whose keys are TXIDs of input transactions, and whose
   *                 values are their associated SPV envelopes.
   * @param {Object} [obj.inputSelection] If Dojo needs to select more inputs
   *                 beyond what you provided in the `inputs` parameter, this
   *                 parameter describes which kinds of inputs can be selected,
   *                 and from where.
   * @param {Boolean} [obj.inputSelection.disable] This is a boolean that, when
   *                 true, will forbid Dojo from adding any additional inputs to
   *                 your transaction, beyond what you specified in the `inputs`
   *                 parameter. Thus, if you have not sufficiently funded the
   *                 transaction yourself, or if `inputs` is empty, you will get
   *                 an error.
   * @param {Number} [obj.inputSelection.maxUnconfirmedChainLength] An integer
   *                 representing the maximum length for any chain of unconfirmed
   *                 parents that a selected input can have. When -1 (the
   *                 default), no maximum is specified. Cannot be zero. When 1,
   *                 indicates that the input must itself be confirmed. When 2,
   *                 input parents must be confirmed. 3 denotes grandparents, 4
   *                 great-grandparents and so forth.
   * @param {Array<TxOutput>} [obj.outputs] External outputs that you will
   *                 include when you create this transaction. These outputs can
   *                 contain custom scripts as specified by recipients. If the
   *                 inputs to the transaction go beyond what is needed to fund
   *                 these outputs (plus the transaction fee), additional
   *                 Dojo-managed UTXOs will be generated to collect the
   *                 remainder (see the `outputGeneration` parameter for more on
   *                 this).
   * @param {Object} [obj.outputGeneration] If Dojo needs to generate additional
   *                 outputs for the transaction beyond what was specified, this
   *                 object describes what kind of outputs to generate, and where
   *                 they should be kept.
   * @param {String} [obj.outputGeneration.method=auto] The method used to
   *                 generate outputs. "auto" selects the amount and types of
   *                 generated outputs based on the selected basket's
   *                 configuration for how many of each type to keep on hand,
   *                 then uses Benford's law to distribute the satoshis across
   *                 them. "single" just uses one output, randomly selected from
   *                 the available types, that contains all the satoshis.
   * @param {Object} [obj.fee] Represents the fee the transaction will pay
   * @param {String} [obj.fee.model=sat/kb] Fee model to use, currently always
   *                 "sat/kb"
   * @param {Number} [obj.fee.value=500] When the fee model is "sat/kb", this is
   *                 the number of satoshis per kilobyte of block space that the
   *                 transaction will pay.
   * @param {Array<String>} [obj.labels] The labels to affix to this transaction
   * @param {String} [obj.note] A numan-readable note describing the transaction
   * @param {String} [obj.recipient] The Paymail handle for the recipient of the transaction
   *
   * @returns {Promise<TransactionTemplate>} The template you need to sign and process
   */
  async createTransaction ({
    inputs,
    inputSelection,
    outputs,
    outputGeneration,
    fee,
    labels,
    note,
    recipient
  }) {
    return await this.createAuthriteRequest('createTransaction', {
      method: 'POST',
      body: { inputs, inputSelection, outputs, outputGeneration, fee, labels, note, recipient }
    })
  }

  /**
   * Returns which BSV network we are using (mainnet or testnet)
   * @param {String} format for the returned string. Either with or without a 'net' suffix.
   * @returns {String} The current BSV network formatted as requested.
   */
  async getNetwork (format = 'default') {
    if (!this.network) {
      const response = await axios.get(`${this.config.dojoURL}/stats`)
      this.network = response.data.network
    }
    return format === 'default' ? this.network : this.network.substring(0, 4)
  }

  /**
   * After a transaction is created (with `createTransaction` or with `getTransactionWithOutputs`) this is used to process the transaction, thereby activating any change outputs and flagging designated inputs as spent
   * @param {Object} obj All parameters are given in an object
   * @param {String} obj.inputs Inputs to spend as part of this transaction
   * @param {String} obj.submittedTransaction The transaction that has been created and signed
   * @param {String} obj.reference The reference number provided by `createTransaction` or `getTransactionWithOutputs`
   * @param {Object} obj.outputMap An object whose keys are derivation prefixes and whose values are corresponding change output numbers from the transaction.
   * @returns {Promise<Object>} An object containing a `note` field with a success message, and `mapiResponses`, for use in creating an SPV Envelope
   */
  async processTransaction ({ inputs, submittedTransaction, reference, outputMap }) {
    try {
      const result = await this.createAuthriteRequest('processTransaction', {
        method: 'POST',
        body: { hex: submittedTransaction, reference, outputMap }
      })
      return result
    } catch (error) {
    // Free up UTXOs since the transaction failed before throwing the error
    // Unless there was a double spend error
      if (reference) {
        try {
          await this.updateTransactionStatus({
            reference,
            status: 'failed'
          })
        } catch (e) { /* ignore, we still need the code below */ }
      }
      // Check if we need to handle a double spend attempt
      if (error.message.toLowerCase().includes('missing inputs')) {
        const woc = new WhatsOnChain(await this.getNetwork())
        const tx = bsv.Transaction(submittedTransaction)

        const spendingTransactions = {}
        for (const [inputTxId, inputEnvelope] of Object.entries(inputs)) {
          // Find the vouts associated with this input transaction
          const vouts = tx.inputs.filter(x => x.prevTxId.toString('hex') === inputTxId).map(x => x.outputIndex)
          let inputTransactionDetails = ''
          try {
            inputTransactionDetails = await axios.post(`https://api.whatsonchain.com/v1/bsv/${await this.getNetwork('nonet')}/ui/tx/decode`, {
              network: await this.getNetwork('nonet'),
              txhex: inputEnvelope.rawTx
            })
          } catch (error) {
            const e = new Error('Could net get transaction(s) responsible for double spend!')
            e.code = 'ERR_WOC_REQUEST_FAILED'
            throw e
          }
          for (const inputVout of vouts) {
            const outputScriptHex = inputTransactionDetails.data.vout[inputVout].scripthash
            // Note: Maybe reduce network calls by calculate output script here:
            // const scriptHash = bsv.crypto.Hash.sha256sha256(Buffer.from(outputScriptHex, 'hex')).toString('hex')
            const spendingCandidates = await woc.historyByScriptHash(outputScriptHex)
            for (const candidate of spendingCandidates) {
              const candidateRawTx = await woc.getRawTxData(candidate.tx_hash)
              const candidateParsedTx = new bsv.Transaction(candidateRawTx)
              const commonInputs = candidateParsedTx.inputs.filter(x => x.prevTxId.toString('hex') === inputTxId && x.outputIndex === inputVout)

              // Check if common inputs were found
              if (commonInputs.length > 0) {
                spendingTransactions[candidate.tx_hash] = 1

                // Set spendable to false for the UTXO that was spent by the tx given as inputTxId
                await this.updateOutpointStatus({ txid: inputTxId, vout: inputVout, spendable: false })
              }
            }
          }
        }

        // Hydrate the envelope data for the spending transactions
        for (const txid of Object.keys(spendingTransactions)) {
          const env = await hashwrap(txid, { network: await this.getNetwork(), taalApiKey: this.taalApiKeys[await this.getNetwork('nonet')] })
          spendingTransactions[txid] = env
        }

        // Format an error message that can be parsed by the calling application.
        const e = new Error('One or more inputs have already been spent! \n Learn how to handle double spends within your apps: https://projectbabbage.com/docs/babbage-sdk/reference/handling-double-spends')
        e.code = 'ERR_DOUBLE_SPEND'
        e.spendingTransactions = spendingTransactions
        throw e
      }
      throw error
    }
  }

  /**
 * Signs and processes all pending transactions, useful when recovering from an
 * error or crash, or on startup. If a transaction fails to process, marks it
 * as failed.
 * @param {Object} obj All parameters are given in an object
 * @param {Function} [obj.onTransactionProcessed] A function called for each processed transaction.
 * @param {Function} [obj.onTransactionFailed] A function called for each failed transaction.
 * @returns {Promise} Resolves once the operation is complete
 */
  async processPendingTransactions ({
    onTransactionProcessed = () => { },
    onTransactionFailed = () => { }
  } = {}) {
    return await _processPendingTransactions({
      authriteClient: this.authriteClient,
      config: this.config,
      onTransactionProcessed,
      onTransactionFailed
    })
  }

  /**
   * Returns a set of transaction outputs that Dojo has tracked
   * @param {Object} obj All parameters are given in an object
   * @param {String} [obj.basket] If provided, indicates which basket the outputs should be selected from.
   * @param {Boolean} [obj.tracked] If provided, only outputs with the corresponding tracked value will be returned (true/false).
   * @param {Boolean} [obj.includeEnvelope] If provided, returns a structure with the SPV envelopes for the UTXOS that have not been spent.
   * @param {Boolean} [obj.spendable] If given as true or false, only outputs that have or have not (respectively) been spent will be returned. If not given, both spent and unspent outputs will be returned.
   * @param {String} [obj.type] If provided, only outputs of the specified type will be returned. If not provided, outputs of all types will be returned.
   * @param {Number} [obj.limit] Provide a limit on the number of outputs that will be returned.
   * @param {Number} [obj.offset] Provide an offset into the list of outputs.
   * @returns {Promise<Array<TransactionOutputDescriptor>>} A set of outputs that match the criteria
   */
  async getTransactionOutputs ({
    basket,
    tracked,
    includeEnvelope = false,
    spendable,
    type,
    limit = 25,
    offset = 0
  } = {}) {
    return await this.createAuthriteRequest('getTransactionOutputs', {
      method: 'POST',
      body: { basket, tracked, includeEnvelope, limit, offset, spendable, type }
    })
  }

  /**
   * Use this endpoint to update the status of a transaction. This is useful for flagging incomplete transactions as aborted or reverting a completed transaction back into a pending status if it never got confirmed. Setting the status to "completed" or "waitingForSenderToSend" will make any selected UTXOs unavailable for spending, while any other status value will free up the UTXOs for use in other transactions.
   * @param {Object} obj All parameters are given in an object
   * @param {String} obj.reference The Dojo reference number for the transaction
   * @param {String} obj.status The new status of the transaction
   * @returns {Promise<Object>} A success object with `status: "success"`
   */
  async updateTransactionStatus ({ reference, status }) {
    return await this.createAuthriteRequest('updateTransactionStatus', {
      method: 'POST',
      body: { status, reference }
    })
  }

  /**
   * This endpoint allows a recipient to submit a transactions that was directly given to them by a sender. Saves the inputs and key derivation information, allowing the UTXOs to be redeemed in the future. Sets the transaction to completed and marks the outputs as spendable.
   *
   * @param {Object} obj All parameters are given in an object
   * @param {string} obj.protocol Specify the transaction submission payment protocol to use. Currently, the only supported protocol is that with BRFC ID "3241645161d8"
   * @param {Object} obj.transaction The transaction envelope to submit, including key derivation information
   * @param {Array} obj.transaction.outputs An array of outputs, each containing `vout`, `satoshis`, `derivationSuffix`, and (optionally), `derivationPrefix`. If a global `derivationPrefix` is used (recommended), output-specific derivation prefixes should be omitted.
   * @param {string} obj.senderIdentityKey Provide the identity key for the person who sent the transaction
   * @param {string} obj.note Human-readable description for the transaction
   * @param {string} [obj.derivationPrefix] A derivation prefix used for all outputs. If provided, derivation prefixes on all outputs are optional.
   *
   * @returns {Promise<Object>} Object containing reference number, status=success, and human-readable note acknowledging the transaction
   */
  async submitDirectTransaction ({
    protocol,
    transaction,
    senderIdentityKey,
    note,
    amount,
    labels,
    derivationPrefix
  }) {
    // A sender identity key is required
    if (typeof senderIdentityKey !== 'string') {
      const e = new Error('A transaction object is required!')
      e.code = 'ERR_MISSING_TRANSACTION'
      throw e
    }

    // A transaction is required as an object
    if (typeof transaction !== 'object') {
      const e = new Error('A transaction object is required!')
      e.code = 'ERR_MISSING_TRANSACTION'
      throw e
    }

    // Transactions need arrays of outputs
    if (!Array.isArray(transaction.outputs)) {
      const e = new Error(
        'The transaction requires an array of output objects!'
      )
      e.code = 'ERR_MISSING_TX_OUTPUTS'
      throw e
    }

    // At least one transaction output is required
    if (transaction.outputs.length < 1) {
      const e = new Error('The transaction needs at least one output!')
      e.code = 'ERR_NO_TX_OUTPUTS_GIVEN'
      throw e
    }

    // All transaction outputs must be objects
    if (!transaction.outputs.every(x => typeof x === 'object')) {
      const e = new Error('All transaction outputs must be objects!')
      e.code = 'ERR_TX_OUTPUT_BAD_TYPE'
      throw e
    }

    // The sender identity key is mapped onto the outputs
    transaction.outputs = transaction.outputs.map(x => {
      if (derivationPrefix) {
        // If provided, the global derivation prefix is also mapped
        return {
          ...x,
          derivationPrefix,
          senderIdentityKey
        }
      } else {
        return {
          ...x,
          senderIdentityKey
        }
      }
    })

    // Each output needs vout, satoshis, derivationPrefix, derivationSuffix
    for (const i in transaction.outputs) {
      const out = transaction.outputs[i]
      if (typeof out.vout !== 'number' || !Number.isInteger(out.vout)) {
        const e = new Error(`Output with index ${i} has invalid vout!`)
        e.outputIndex = i
        e.code = 'ERR_BAD_OUTPUT_VOUT'
        throw e
      }
      if (
        typeof out.satoshis !== 'number' ||
        !Number.isInteger(out.satoshis) ||
        out.satoshis < 1 ||
        out.satoshis > 21e8
      ) {
        const e = new Error(`Output with index ${i} has invalid satoshis!`)
        e.outputIndex = i
        e.code = 'ERR_BAD_OUTPUT_SATOSHIS'
        throw e
      }

      if (typeof out.derivationPrefix !== 'string') {
        const e = new Error(
          `Output with index ${i} has invalid derivationPrefix. Use the same derivationPrefix for all outputs.`
        )
        e.outputIndex = i
        e.code = 'ERR_BAD_OUTPUT_PREFIX'
        throw e
      }

      if (typeof out.derivationSuffix !== 'string') {
        const e = new Error(
          `Output with index ${i} has invalid derivationSuffix.`
        )
        e.outputIndex = i
        e.code = 'ERR_BAD_OUTPUT_SUFFIX'
        throw e
      }
    }

    // All derivationPrefixes must be for the same payment
    if (
      !transaction.outputs.every(
        x => x.derivationPrefix === transaction.outputs[0].derivationPrefix
      )
    ) {
      const e = new Error(
        'Every output in a given payment must use the same derivation prefix.'
      )
      e.code = 'ERR_INCONSISTENT_DERIVATION_PREFIX'
      throw e
    }

    // eslint-disable-next-line
    return new Promise(async (resolve, reject) => {
      try {
        const processed = await _processIncomingTransaction({
          authriteClient: this.authriteClient,
          config: this.config,
          updateStatus: false,
          incomingTransaction: {
            inputs: transaction.inputs,
            outputs: transaction.outputs,
            rawTransaction: transaction.rawTx
          },
          onTransactionFailed: ({ error }) => { reject(error) }
        })
        if (!processed) {
          return // Error is rejected, above.
        }
        if (typeof amount === 'number') {
          if (processed.amount !== amount) {
            const e = new Error(
              `Transaction amount is not correct! Expected: ${amount}, but got: ${processed.amount}. Transaction was not broadcast, processed or received.` // "Transaction was not broadcast, processed or received" on every error?
            )
            e.code = 'ERR_TX_BAD_AMOUNT'
            reject(e)
            return
          }
        }
        const submitResult = await this.createAuthriteRequest(
          'submitDirectTransaction',
          {
            method: 'POST',
            body: {
              protocol,
              derivationPrefix: transaction.outputs[0].derivationPrefix,
              transaction: { // Transform transaction outputs to required format
                ...transaction,
                outputs: Object.fromEntries(transaction.outputs.map(out => ([
                  out.vout,
                  {
                    suffix: out.derivationSuffix,
                    basket: out.basket
                  }
                ])))
              },
              labels,
              senderIdentityKey,
              note
            }
          }
        )
        resolve(submitResult)
      } catch (e) {
        reject(e)
      }
    })
  }

  /**
   * Verifies an incoming Paymail transaction (deprecated, use submitDirectTransaction)
   * @param {Object} obj
   * @param {string} obj.senderPaymail
   * @param {string} obj.senderIdentityKey
   * @param {string} obj.referenceNumber
   * @param {string} obj.description
   * @returns {Promise<Boolean>} A success boolean status
   */
  async verifyIncomingTransaction ({
    senderPaymail,
    senderIdentityKey,
    referenceNumber,
    description,
    amount
  }) {
    // Check if the referenced transaction is pending
    const [transaction] = await this.getPendingTransactions({ referenceNumber })
    if (transaction && transaction.status === 'unprocessed' && transaction.amount === amount && transaction.senderPaymail === senderPaymail) {
      // Validate the sender paymail and identityKey for each output
      const outputs = transaction.outputs
      for (let i = 0; i < outputs.length; i++) {
        if (outputs[i].senderIdentityKey !== senderIdentityKey) {
          return false
        }
      }
      // Process the pending transaction
      return new Promise((resolve, reject) => {
        try {
          _processIncomingTransaction({
            authriteClient: this.authriteClient,
            config: this.config,
            incomingTransaction: transaction,
            onTransactionProcessed: () => {
              resolve(true)
            },
            onTransactionFailed: () => {
              resolve(false)
            }
          })
        } catch (e) {
          reject(e)
        }
      })
    }
    // Verify the transaction was successfully processed
    // TODO: Check senderIdentityKey before returning true
    const [finalResult] = (await this.getTransactions({ referenceNumber, limit: 1 })).transactions
    if (finalResult.status === 'completed' && finalResult.senderPaymail === senderPaymail && finalResult.note === description && finalResult.amount === amount) {
      // const outputs = finalResult.outputs
      // for (let i = 0; i < outputs.length; i++) {
      //   if (outputs[i].senderIdentityKey !== senderIdentityKey) {
      //     return false
      //   }
      // }
      return true
    }
    return false
  }

  /**
   * Use this endpoint to update the status of one of your outputs, given as the TXID of a transaction and the vout (output index) in that transaction. This is useful for flagging transaction outpoints as spent if they were inadvertantly broadcasted or used without properly submitting them to the Dojo, or to undo the spending of an output if it was never actually spent.
   * @param {Object} obj All parameters are given in an object
   * @param {String} obj.txid The TXID of the transaction that created the output
   * @param {Number} obj.vout The index of the output in the transaction
   * @param {Boolean} obj.spendable The true spendability status of this outpoint
   * @returns {Promise<Object>} A success object with `status: "success"`
   */
  async updateOutpointStatus ({ txid, vout, spendable }) {
    return await this.createAuthriteRequest('updateOutpointStatus', {
      method: 'POST',
      body: { txid, vout, spendable }
    })
  }

  /**
   * Use this endpoint to store an incoming certificate.
   * @param {Object} obj All parameters are given in an object
   * @param {Object} obj.certificate The certificate object to save
   * @returns {Promise<Object>} A success object with `status: "success"`
   */
  async saveCertificate ({ certificate }) {
    return await this.createAuthriteRequest('saveCertificate', {
      method: 'POST',
      body: { certificate }
    })
  }

  /**
   * Use this endpoint to retrieve certificates.
   * @param {Object} obj All parameters are given in an object
   * @param {Array} obj.certifiers The certifiers to filter certificates by
   * @param {Object} obj.types The certificate types to filter certificates by
   * @returns {Promise<Object>} A success object with `status: "success"` and any found certificates
   */
  async findCertificates ({ certifiers, types }) {
    return await this.createAuthriteRequest('findCertificates', {
      method: 'POST',
      body: { certifiers, types }
    })
  }
}

module.exports = Ninja
