const bsv = require('babbage-bsv')
const { getPaymentPrivateKey, getPaymentAddress } = require('sendover')
const makeCreateAuthriteRequest = require('./utils/makeCreateAuthriteRequest')
const processIncomingTransaction = require('./utils/processIncomingTransaction')

module.exports = async ({
  authriteClient,
  config,
  onTransactionProcessed = () => { },
  onTransactionFailed = () => { }
} = {}) => {
  const createAuthriteRequest = makeCreateAuthriteRequest({
    authriteClient,
    config
  })
  const pending = await createAuthriteRequest('getPendingTransactions', {
    method: 'POST',
    body: {}
  }
  )
  await Promise.all(pending.map(async pendingTx => {
    const {
      inputs,
      outputs: txOutputs,
      referenceNumber,
      isOutgoing
    } = pendingTx
    // Pending outgoing transactions are signed and processed with processTransaction
    if (isOutgoing) {
      let hex
      try {
        const tx = new bsv.Transaction()
        const outputMap = {}
        txOutputs.forEach((out, i) => {
          if (out.providedBy === 'dojo' && out.purpose === 'change') {
            // Derive a change output script
            // Get derivation invoice data
            const derivationSuffix = out.instructions.derivationSuffix
            outputMap[derivationSuffix] = i
            const invoiceNumber = out.paymailHandle
              ? `2-3241645161d8-${out.paymailHandle} ${out.derivationPrefix} ${derivationSuffix}`
              : `2-3241645161d8-${out.derivationPrefix} ${derivationSuffix}`
            // Derive the public key used for creating the output script
            const derivedAddress = getPaymentAddress({
              senderPublicKey: authriteClient.clientPublicKey,
              recipientPrivateKey: authriteClient.clientPrivateKey,
              invoiceNumber,
              returnType: 'address'
            })
            // Create an output script that can only be unlocked with the corresponding derived private key
            tx.addOutput(new bsv.Transaction.Output({
              script: new bsv.Script(
                bsv.Script.fromAddress(derivedAddress)
              ),
              satoshis: out.satoshis
            }))
          } else {
            tx.addOutput(new bsv.Transaction.Output({
              script: new bsv.Script(out.script),
              satoshis: out.satoshis
            }))
          }
        })

        const getIndex = o => {
          if (typeof o === 'object') {
            return o.index
          } else {
            return o
          }
        }

        // Add inputs, and sum input amounts
        for (const [inputTXID, input] of Object.entries(inputs)) {
          const t = new bsv.Transaction(input.rawTx)
          for (const o of input.outputsToRedeem) {
            tx.from(bsv.Transaction.UnspentOutput({
              txid: inputTXID,
              outputIndex: getIndex(o),
              // scruptPubKey a.k.a. lockingScript or outputScript
              // (whereas scriptSig a.k.a. unlockingScript or inputScript)
              scriptPubKey: t.outputs[getIndex(o)].script,
              satoshis: t.outputs[getIndex(o)].satoshis
            }))
            if ( // All foreign input scripts are added
              inputs[t.id] &&
        inputs[t.id].outputsToRedeem.some(x => x.index === getIndex(o))
            ) {
              tx.inputs[tx.inputs.length - 1].setScript(bsv.Script.fromHex(
                inputs[t.id].outputsToRedeem[getIndex(o)].unlockingScript
              ))
            }
          }
        }

        //  Sign inputs using type42 derived key
        for (const input of Object.values(inputs)) {
          for (const o of input.outputsToRedeem) {
            if (input.instructions && input.instructions[getIndex(o)]) {
              const instructions = input.instructions[getIndex(o)]

              // Make sure the transaction type is supported
              if (instructions.type !== 'P2PKH') {
                throw new Error(
            `The server is specifying ${instructions.type}, which is not a supported unlocking script type. Supported type: P2PKH`
                )
              }

              // Get derivation invoice data
              const paymailHandle = instructions.paymailHandle
              const derivationPrefix = instructions.derivationPrefix
              const derivationSuffix = instructions.derivationSuffix

              const invoiceNumber = paymailHandle
                ? `2-3241645161d8-${paymailHandle} ${derivationPrefix} ${derivationSuffix}`
                : `2-3241645161d8-${derivationPrefix} ${derivationSuffix}`
              // Derive the key used to unlock funds
              const derivedPrivateKey = getPaymentPrivateKey({
                recipientPrivateKey: authriteClient.clientPrivateKey,
                senderPublicKey: instructions.senderIdentityKey,
                invoiceNumber,
                returnType: 'hex'
              })
              tx.sign(bsv.PrivateKey.fromHex(derivedPrivateKey))
            }
          }
        }
        hex = tx.uncheckedSerialize()
        const processResult = await createAuthriteRequest(
          'processTransaction', {
            method: 'POST',
            body: {
              hex,
              reference: referenceNumber,
              outputMap
            }
          })
        try {
          await onTransactionProcessed({
            inputs,
            isOutgoing,
            hex,
            reference: referenceNumber,
            outputMap,
            txid: tx.id,
            ...processResult
          })
        } catch (e) {
          console.error('onTransactionProcessed callback threw', e)
        }
      } catch (e) {
        console.error('Setting transaction as failed due to error with processing pending', e)
        try {
          await onTransactionFailed({
            error: e,
            inputs,
            isOutgoing,
            reference: referenceNumber
          })
        } catch (e) {
          console.error('onTransactionFailed callback threw', e)
        }
        try {
          await createAuthriteRequest('updateTransactionStatus', {
            method: 'POST',
            body: { status: 'failed', reference: referenceNumber }
          })
        } catch (e) { /* ignore, so that we can still deal with other TXs */ }
      }

    // pending incoming transactions are first validated and then their status is updated to be completed or failed
    } else {
      await processIncomingTransaction({
        authriteClient,
        config,
        incomingTransaction: pendingTx,
        onTransactionProcessed,
        onTransactionFailed
      })
    }
  }))
}
