const bsv = require('babbage-bsv')
const { getPaymentPrivateKey } = require('sendover')
const makeCreateAuthriteRequest = require('./makeCreateAuthriteRequest')

module.exports = async ({
  authriteClient,
  config,
  incomingTransaction,
  updateStatus = true,
  onTransactionProcessed = () => { },
  onTransactionFailed = () => { }
} = {}) => {
  const createAuthriteRequest = makeCreateAuthriteRequest({
    authriteClient,
    config
  })
  const {
    inputs,
    outputs: txOutputs,
    referenceNumber,
    rawTransaction
  } = incomingTransaction
  try {
    const tx = new bsv.Transaction(rawTransaction)
    let senderIdentityKey, derivationPrefix
    let amount = 0
    // Derive and check output scripts
    for (const out of txOutputs) {
      const derivationSuffix = out.derivationSuffix
      const invoiceNumber = out.paymailHandle
        ? `2-3241645161d8-${out.paymailHandle} ${out.derivationPrefix} ${derivationSuffix}`
        : `2-3241645161d8-${out.derivationPrefix} ${derivationSuffix}`
      // Derive the public key used for creating the output script
      const derivedPrivateKey = getPaymentPrivateKey({
        senderPublicKey: out.senderIdentityKey,
        recipientPrivateKey: authriteClient.clientPrivateKey,
        invoiceNumber,
        returnType: 'wif'
      })
      const derivedAddress = bsv.Address.fromPrivateKey(
        bsv.PrivateKey.fromWIF(derivedPrivateKey)
      )
      const expectedScript = new bsv.Script(
        bsv.Script.fromAddress(derivedAddress)
      ).toHex()
      if (tx.outputs[out.vout].script.toHex() !== expectedScript) {
        throw new Error('Transaction output script invalid')
      }
      if (tx.outputs[out.vout].satoshis !== parseInt(out.satoshis)) {
        throw new Error('Transaction output amount invalid')
      }
      senderIdentityKey = out.senderIdentityKey
      derivationPrefix = out.derivationPrefix
      amount = parseInt(amount) + parseInt(out.satoshis)
    }
    if (updateStatus) {
      await createAuthriteRequest('updateTransactionStatus', {
        method: 'POST',
        body: { status: 'completed', reference: referenceNumber }
      })
    }
    try {
      const returnValue = {
        inputs,
        isOutgoing: false,
        hex: rawTransaction,
        outputs: txOutputs,
        senderIdentityKey,
        derivationPrefix,
        txid: tx.id,
        reference: referenceNumber,
        amount
      }
      await onTransactionProcessed(returnValue)
      return returnValue
    } catch (e) {
      console.error('onTransactionProcessed callback threw', e)
      return true
    }
  } catch (e) {
    console.error('Received invalid incoming transaction, marking it as failed', e)
    if (updateStatus) {
      await createAuthriteRequest('updateTransactionStatus', {
        method: 'POST',
        body: { status: 'failed', reference: referenceNumber }
      })
    }
    try {
      await onTransactionFailed({
        error: e,
        inputs,
        isOutgoing: false,
        reference: referenceNumber
      })
      return false
    } catch (e) {
      console.error('onTransactionFailed callback threw', e)
      return false
    }
  }
}
